//基本的なクラス定義の例
class Car(){
  //メソッド定義
  fun start(){
    println("発進")
  }
  fun stop(){
    println("停止")
  }

}

//プライマリコンストラクタ。name引数で名のみを指定
class Person(name: String){
  //姓、名プロパティ
  var firstName: String
  var lastName: String

  //プライマリコンストラクタの処理内容はinitで記述
  init{
    this.firstName = name //プライマリコンストラクタの引数をプロパティに代入
    this.lastName = "" //姓は未指定とする
  }

  //セカンダリコンストラクタ。姓、名両方を指定
  //this()でプライマリコンストラクタを呼び出す
  constructor(firstName: String, lastName: String): this(firstName){
    this.lastName = lastName //姓に2番目の引数を代入する
  }

  fun print(){
    println("お名前: ${firstName} ${lastName}")
  }
}

//プライマリコンストラクタでプロパティを定義。デフォルト値も設定
class Person2(val firstName: String, val lastName: String = ""){
  fun print(){
    //プライマリコンストラクタで定義されたプロパティを参照
    println("お名前: ${firstName} ${lastName}")
  }
}

//データクラス
data class PersonByDataClass(var firstName: String, var lastName: String = "")

//コンパニオンオブジェクトの使用例
class CompanionSample(val firstName: String, val lastName: String = ""){
  companion object{
    //helloメソッドを定義
    fun hello(name: String){
      println("こんにちは $name さん")
    }
    //プロパティを定義
    val PI = 3.141592653589793
  }
}

object SingletonSample{}

//設定ファイル読み書きオブジェクト。常に実体は1つ
object Config{
  //設定ファイル読み書きダミー実装用Map
  val map = mutableMapOf<String, String>()
  fun read(key: String): String{
    return map[key] ?: ""
  }
  fun write(key: String, value: String){
    map[key] = value
  }
}

//プロパティを持つクラス定義の例
class Counter(){
  //書き換え可能なプロパティ
  var counter: Int = 0
  //書き換え不可能なプロパティ。ここではデータ型を省略して推論させている
  val maxValue = 10
  fun countUp(){
    //プロパティの値に1を加算
    counter += 1
  }
  fun print(){
    println("現在のカウント値: ${counter}")
  }
}

//時計クラスの定義
class Clock(){
  //時・分・秒のプロパティ
  var hour: Int = 0
  var minute: Int = 0
  var second: Int = 0
  //00:00:00から何秒経過したかを取得・設定するプロパティ
  var pastSeconds: Int
     get(){  //getキーワードで取得の際のロジックを記述
        return ((hour * 60) + minute) * 60 + second
     }
     set(value){
        hour = value / 3600
        minute = (value % 3600) / 60
        second = value % 60
     }

  fun print(){ //現在のプロパティの値を表示
    //pastSecondsプロパティはgetterロジックを実行
    println("${hour}時${minute}分${second}秒, 0時から${pastSeconds}秒経過")
  }
}

fun main(args:Array<String>) {

    //クラスの使用例
    //newではなく、クラス名をメソッドのように使ってインスタンスを作る
    val car = Car()
    //メソッド呼び出しなどはJavaと同様
    car.start()
    car.stop()

    //プロパティの使用例
    val counter = Counter()
    counter.print() //「現在のカウント値: 0」が出力される
    counter.countUp()
    counter.countUp()
    counter.print() //「現在のカウント値: 2」が出力される

    val clock = Clock()
    //普通にプロパティに値を設定。ここはフィールドと同様
    clock.hour = 1
    clock.minute = 10
    clock.second = 5
    clock.print() // 「1時10分5秒, 0時から4205秒経過」が出力される
    //pastSecondsプロパティに値を設定。setterロジックを実行
    clock.pastSeconds = 4101
    clock.print() //「1時8分21秒, 0時から4101秒経過」が出力される

    //プライマリコンストラクタの呼び出し例
    val person = Person("Doi")
    person.print() //結果: 「お名前: Doi」

    //セカンダリコンストラクタの呼び出し例
    val person2 = Person("Tsuyoshi", "Doi")
    person2.print() //結果: 「お名前: Tsuyoshi Doi」

    //第2引数を省略してプライマリコンストラクタを呼び出す
    val person3 = Person2("Doi")
    person3.print() //結果: 「お名前: Doi」

    //第2引数を指定してプライマリコンストラクタを呼び出す
    val person4 = Person2("Tsuyoshi", "Doi")
    person4.print() //結果: 「お名前: Tsuyoshi Doi」

    //データクラスを定義
    val person5 = PersonByDataClass("Tsuyoshi", "Doi")
    //データクラスには自動でtoStringメソッドが定義される
    println(person5) //結果: 「PersonByDataClass(firstName=Tsuyoshi, lastName=Doi)」

    //データクラスはインスタンスをコピーするcopyメソッドを持つ
    val person6 = person5.copy()
    person6.firstName = "TsuyoTsuyo"
    println(person6) //結果: 「PersonByDataClass(firstName=TsuyoTsuyo, lastName=Doi)」

    //データクラスはN番目に定義されたプロパティを返すcomponentNメソッドを持つ
    println(person6.component2()) //結果: 「Doi」←2番目のプロパティであるlastNameの値

    //コンパニオンオブジェクトで定義されたメソッドをstaticメソッドのように呼び出し
    CompanionSample.hello("土井") //結果: 「こんにちは 土井 さん」
    //コンパニオンオブジェクトで定義されたプロパティをstatic変数のように呼び出し
    println(CompanionSample.PI) //結果: 「3.141592653589793」

    //シングルトンオブジェクトの使用例
    //Configはクラスではなくオブジェクトなので、そのままメソッドを呼び出せる
    Config.write("test1", "value1") //設定書込
    val value = Config.read("test1") //設定読込
    print(value)
}
